#ifdef notes_go_here

BOOL EscapeCommFunction(
  HANDLE hFile,  // handle to communications device
  DWORD dwFunc   // extended function to perform
);

Value Meaning 
CLRDTR Clears the DTR (data-terminal-ready) signal. 
CLRRTS Clears the RTS (request-to-send) signal. 
SETDTR Sends the DTR (data-terminal-ready) signal. 
SETRTS Sends the RTS (request-to-send) signal. 
SETXOFF Causes transmission to act as if an XOFF character has been received. 
SETXON Causes transmission to act as if an XON character has been received. 
SETBREAK Suspends character transmission and places the transmission line in a break state until the ClearCommBreak function is called (or EscapeCommFunction is called with the CLRBREAK extended function code). The SETBREAK extended function code is identical to the SetCommBreak function. Note that this extended function does not flush data that has not been transmitted. 
CLRBREAK Restores character transmission and places the transmission line in a nonbreak state. The CLRBREAK extended function code is identical to the ClearCommBreak function. 

#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>

static HANDLE com_port = INVALID_HANDLE_VALUE;

static void
usage() {
   printf("rs232 - control an rs232 port\n\
usage: rs232\n\
\n\
No options, use help command once prompted for input to get\n\
list of interactive commands\n");
}

static void help_cmd();

static void
splain_error(char * msg, DWORD errcode) {
   // PUT CODE HERE - use format message
   printf("%s failed, reason %#x\n", msg, errcode);
}

static void
quit_cmd() {
   exit(0);
}

static void
open_cmd() {
   DWORD dummy1;
   COMSTAT dummy2;
   char * port_name = strtok(NULL, " \t\n\r");
   char * port_parm = strtok(NULL, "\n\r");
   if (port_parm != NULL) {
      while (isspace(*port_parm)) {
         ++port_parm;
      }
      if (*port_parm == '\0') {
         port_parm = NULL;
      }
   }
   if (port_parm == NULL) {
      port_parm = "baud=9600 parity=N data=8 stop=1";
   }

   if (com_port != INVALID_HANDLE_VALUE) {
      printf("There is already a com port open\n");
      return;
   }

   COMMTIMEOUTS tmouts;
   memset((void *)&tmouts, 0, sizeof(tmouts));
   tmouts.ReadIntervalTimeout = 1;

   com_port = CreateFile(port_name,
                         GENERIC_READ | GENERIC_WRITE,
                         0,
                         NULL,
                         OPEN_EXISTING,
                         FILE_FLAG_OVERLAPPED,
                         NULL);

   if (com_port == INVALID_HANDLE_VALUE) {
      DWORD whyfor = GetLastError();
      splain_error("CreateFile", whyfor);
      return;
   }
   PurgeComm(com_port,
      PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
   ClearCommError(com_port, &dummy1, &dummy2);
   if (! SetCommMask(com_port, EV_ERR)) {
      DWORD whyfor = GetLastError();
      splain_error("SetCommMask", whyfor);
      CloseHandle(com_port);
      com_port = INVALID_HANDLE_VALUE;
      return;
   }
   if (! SetCommTimeouts(com_port, &tmouts)) {
      DWORD whyfor = GetLastError();
      splain_error("SetCommTimeouts", whyfor);
      CloseHandle(com_port);
      com_port = INVALID_HANDLE_VALUE;
      return;
   }

   union {
      DCB dcb;
      DCB extra_space[3];
   };
   memset((void *)&dcb, 0, sizeof(dcb));
   dcb.DCBlength = sizeof(dcb);
   dcb.BaudRate = CBR_9600;
   dcb.fBinary = TRUE;
   dcb.fParity = FALSE;
   dcb.fOutxCtsFlow = FALSE;
   dcb.fOutxDsrFlow = FALSE;
   dcb.fDsrSensitivity = FALSE;
   dcb.fOutX = FALSE;
   dcb.fInX = FALSE;
   dcb.fErrorChar = FALSE;
   dcb.fNull = FALSE;
   dcb.fAbortOnError = FALSE;
   if (! BuildCommDCB(port_parm, &dcb)) {
      DWORD whyfor = GetLastError();
      splain_error("BuildCommDCB", whyfor);
      CloseHandle(com_port);
      com_port = INVALID_HANDLE_VALUE;
      return;
   }
   if (! SetCommState(com_port, &dcb)) {
      DWORD whyfor = GetLastError();
      splain_error("SetCommState", whyfor);
      CloseHandle(com_port);
      com_port = INVALID_HANDLE_VALUE;
      return;
   }
   ClearCommBreak(com_port);
   PurgeComm(com_port,
      PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
   ClearCommError(com_port, &dummy1, &dummy2);
   if (! SetupComm(com_port, 64*1024, 4*1024)) {
      DWORD whyfor = GetLastError();
      splain_error("SetupComm", whyfor);
      CloseHandle(com_port);
      com_port = INVALID_HANDLE_VALUE;
      return;
   }
}

static void
close_cmd() {
   CloseHandle(com_port);
   com_port = INVALID_HANDLE_VALUE;
}

static void
do_escape(DWORD parm) {
   if (! EscapeCommFunction(com_port, parm)) {
      DWORD whyfor = GetLastError();
      splain_error("EscapeCommFunction", whyfor);
   }
}

static void
clrdtr_cmd() {
   do_escape(CLRDTR);
}

static void
clrrts_cmd() {
   do_escape(CLRRTS);
}

static void
setdtr_cmd() {
   do_escape(SETDTR);
}

static void
setrts_cmd() {
   do_escape(SETRTS);
}

static void
setbreak_cmd() {
   do_escape(SETBREAK);
}

static void
clrbreak_cmd() {
   do_escape(CLRBREAK);
}

typedef void voidfunc();

struct cmdtab_struct {
   char * name;
   voidfunc * funcptr;
};

struct cmdtab_struct cmdtab [] = {
   {"quit", quit_cmd},
   {"help", help_cmd},
   {"open", open_cmd},
   {"close", close_cmd},
   {"clrdtr", clrdtr_cmd},
   {"clrrts", clrrts_cmd},
   {"setdtr", setdtr_cmd},
   {"setrts", setrts_cmd},
   {"setbreak", setbreak_cmd},
   {"clrbreak", clrbreak_cmd},
   {NULL, NULL}
};

static void
help_cmd() {
   printf("Command names:\n");
   for (int i = 0; cmdtab[i].name != NULL; ++i) {
      printf("   %s\n", cmdtab[i].name);
   }
}

static void
do_one_command(char * inbuf) {
   char * token = strtok(inbuf, " \t\n\r");
   int i = 0;
   while (cmdtab[i].name != NULL) {
      if (strcmp(token, cmdtab[i].name) == 0) {
         (*cmdtab[i].funcptr)();
         return;
      }
      ++i;
   }
}

int
main(int argc, char ** argv) {
   char inbuf[128];

   if (argc > 1) {
      usage();
      return 0;
   }
   while (1) {
      printf("rs232> ");
      fflush(stdout);
      if (fgets(inbuf, sizeof(inbuf)-2, stdin) == NULL) {
         break;
      }
      inbuf[sizeof(inbuf)-2] = '\n';
      inbuf[sizeof(inbuf)-1] = '\0';
      do_one_command(inbuf);
   }
   return 0;
}
